@@ -1,10 +1,12 @@
 import random
 import typing
 import urllib.parse
+import tempfile
+
 from datetime import datetime
 from timeit import default_timer as timer
 
-from sqlalchemy import Column, Table, MetaData, create_engine
+from sqlalchemy import Column, Table, MetaData, create_engine, DateTime
 
 from bakplane.bakplane_pb2 import (
     ResolveResourcePathResponse,
@@ -19,39 +21,48 @@ from bakplane.extensions.base import (
     ReadRequest,
     ReadRequestBuildingBlocks,
     DataStoreHelper,
+    INFINITY_PROXY_DATE
 )
 from bakplane.extensions.spark.utils import (
     to_spark_save_mode,
     timestamp_to_date_string,
     datetime_to_date_string,
-    timestamp_to_datetime,
 )
 from bakplane.utils import to_execution_statistics
 
 
 class JDBCStoreHelper(DataStoreHelper):
     def __init__(
-        self,
-        user: str,
-        password: str,
-        host: str,
-        port: int,
-        database: str,
-        schema_name: str,
-        driver: str,
+            self,
+            user: str,
+            password: str,
+            host: str,
+            port: int,
+            database: str,
+            schema_name: str,
+            driver: str,
+            table_prefixes: typing.List[str] = None
     ):
         self.metadata = MetaData()
         self.engine = create_engine(
-            f"{driver}://{user}:{password}@{host}:{port}/{database}"
+            f"{driver}://{user}:{password}@{host}:{port}/{database}",
+            executemany_values_page_size=10000,
+            executemany_batch_page_size=500,
+            executemany_mode='batch',
         )
         self.schema_name = schema_name
+        self.table_prefixes = table_prefixes
 
     def create_temporary_table(
-        self,
-        prefix: str,
-        columns: typing.List[Column],
-        rows: typing.List[typing.Dict[str, typing.Any]],
+            self,
+            prefix: str,
+            columns: typing.List[Column],
+            rows: typing.List[typing.Dict[str, typing.Any]],
     ) -> Table:
+
+        for c in columns:
+            c.table = None
+
         tbl = Table(
             prefix
             + "_"
@@ -63,23 +74,97 @@ class JDBCStoreHelper(DataStoreHelper):
             schema=self.schema_name,
         )
 
-        self.metadata.create_all(self.engine)
+        tbl.create(bind=self.engine)
         self.engine.execute(tbl.insert(), rows)
 
         return tbl
 
 
 def __build_read_all_query(
-    r: ReadAllRequest, res: ResolveResourcePathResponse
+        r: ReadAllRequest, res: ResolveResourcePathResponse
 ) -> str:
     return f'select {", ".join(r.columns)} from {res.context["schema_name"]}.{res.context["table_name"]}'
 
 
+def __add_effective_dating_constraint(blocks: ReadRequestBuildingBlocks, res: ResolveResourceIntentResponse):
+    a = blocks.asset_table
+
+    if blocks.effective_dating_table is not None:
+        blocks.constraints.append(
+            blocks.effective_dating_table.c.dt >= a.c.effective_start_dt
+        )
+
+        blocks.constraints.append(
+            blocks.effective_dating_table.c.dt < a.c.effective_end_dt
+        )
+    else:
+        blocks.constraints.append(
+            a.c.effective_start_dt
+            <= timestamp_to_date_string(
+                res.effective_dating.end_dt.seconds
+            )
+        )
+        blocks.constraints.append(
+            a.c.effective_end_dt
+            >= timestamp_to_date_string(
+                res.effective_dating.start_dt.seconds
+            )
+        )
+
+
+def __add_valid_dating_constraint(blocks: ReadRequestBuildingBlocks, res: ResolveResourceIntentResponse):
+    a = blocks.asset_table
+
+    if blocks.valid_dating_table is not None:
+        blocks.constraints.append(
+            blocks.valid_dating_table.c.dt >= a.c.valid_start_dt
+        )
+
+        blocks.constraints.append(
+            blocks.valid_dating_table.c.dt < a.c.valid_end_dt
+        )
+
+
+def __populate_building_blocks_effective_dating(helper: DataStoreHelper, blocks: ReadRequestBuildingBlocks, r: ReadRequest):
+    if isinstance(r.effective_dating, list):
+        blocks.effective_dating_table = helper.create_temporary_table('t_dating', [Column('dt', DateTime, nullable=False)], [{'dt': x} for x in r.get_resolved_effective_dating()])
+    elif isinstance(r.effective_dating, tuple):
+        start, end = r.effective_dating
+
+        blocks.effective_start_dt = start
+        blocks.effective_end_dt = end
+    elif isinstance(r.effective_dating, datetime):
+        blocks.effective_start_dt = r.effective_dating
+        blocks.effective_end_dt = INFINITY_PROXY_DATE
+    else:
+        raise ValueError('Effective dating has unknown type.')
+
+
+def __populate_building_blocks_valid_dating(helper: DataStoreHelper, blocks: ReadRequestBuildingBlocks, r: ReadRequest):
+
+    if r.valid_dating is None:
+        return
+
+    if isinstance(r.valid_dating, list):
+        blocks.valid_dating_table = helper.create_temporary_table('t_dating', [Column('dt', DateTime, nullable=False)], [{'dt': x} for x in r.get_resolved_valid_dating()])
+    elif isinstance(r.valid_dating, tuple):
+        start, end = r.valid_dating
+
+        blocks.valid_start_dt = start
+        blocks.valid_end_dt = end
+
+    elif isinstance(r.valid_dating, datetime):
+        blocks.valid_start_dt = r.valid_dating
+        blocks.valid_end_dt = INFINITY_PROXY_DATE
+    else:
+        raise ValueError('Valid dating has unknown type.')
+
+
 def __build_read_query(
-    r: ReadRequest,
-    res: ResolveResourceIntentResponse,
-    dialect,
-    store_helper: DataStoreHelper,
+        r: ReadRequest,
+        res: ResolveResourceIntentResponse,
+        dialect,
+        store_helper: DataStoreHelper,
 ) -> str:
     from sqlalchemy import (
         Table,
@@ -127,19 +212,16 @@ def __build_read_query(
         mapping_table=m,
         asset_table=a,
         selectable=m.join(a, and_(*clauses)),
-        effective_start_dt=timestamp_to_datetime(
-            res.effective_dating.start_dt.seconds
-        ),
-        effective_end_dt=timestamp_to_datetime(
-            res.effective_dating.end_dt.seconds
-        ),
-        valid_start_dt=r.valid_start_dt,
-        valid_end_dt=r.valid_end_dt,
-        knowledge_start_dt=r.knowledge_start_dt,
-        knowledge_end_dt=r.knowledge_end_dt,
-        constraints=[],
+        knowledge_dt=r.knowledge_dt,
+        constraints=[and_(*[
+            datetime_to_date_string(r.knowledge_dt) >= a.c.knowledge_start_dt,
+            datetime_to_date_string(r.knowledge_dt) < a.c.knowledge_end_dt,
+        ])],
     )
 
+    __populate_building_blocks_effective_dating(store_helper, blocks, r)
+    __populate_building_blocks_valid_dating(store_helper, blocks, r)
+
     if (r.columns is None or len(r.columns) <= 0) and r.query_builder is None:
         raise RuntimeError("You must provide a set of columns.")
 
@@ -147,30 +229,24 @@ def __build_read_query(
         res = r.query_builder(blocks)
     else:
         blocks.columns = [a.c.get(x) for x in r.columns if x in a.c]
-        blocks.constraints = []
 
         if r.additional_constraints is not None:
             blocks.constraints.append(r.additional_constraints(blocks))
 
+        __add_valid_dating_constraint(blocks, res)
+
         if r.effective_dating_hint is None:
+
             blocks.constraints.append(
                 a.c.effective_start_dt <= m.c.effective_end_dt
             )
+
             blocks.constraints.append(
                 a.c.effective_end_dt >= m.c.effective_start_dt
             )
-            blocks.constraints.append(
-                a.c.effective_start_dt
-                <= timestamp_to_date_string(
-                    res.effective_dating.end_dt.seconds
-                )
-            )
-            blocks.constraints.append(
-                a.c.effective_end_dt
-                >= timestamp_to_date_string(
-                    res.effective_dating.start_dt.seconds
-                )
-            )
+
+            __add_effective_dating_constraint(blocks, res)
+
         else:
             if isinstance(r.effective_dating_hint, tuple):
                 s = datetime_to_date_string(
@@ -199,19 +275,19 @@ def __build_read_query(
 
         res = (
             select(blocks.columns)
-            .select_from(blocks.selectable)
-            .where(and_(*blocks.constraints))
+                .select_from(blocks.selectable)
+                .where(and_(*blocks.constraints))
         )
 
     query = str(
-        res.compile(dialect=dialect, compile_kwargs={"literal_binds": True},)
+        res.compile(dialect=dialect, compile_kwargs={"literal_binds": True}, )
     )
 
     return query
 
 
 def write_fn(
-    r: WriteRequest, res: ResolveResourcePathResponse, driver: str, url: str
+        r: WriteRequest, res: ResolveResourcePathResponse, driver: str, url: str
 ) -> WriteResponse:
     start = timer()
 
@@ -233,18 +309,18 @@ def write_fn(
 
 
 def __read_query(
-    spark, url: str, driver: str, query: str, user: str, password: str
+        spark, url: str, driver: str, query: str, user: str, password: str
 ):
     start = timer()
 
     df = (
         spark.read.format("jdbc")
-        .option("url", url)
-        .option("query", query)
-        .option("user", user)
-        .option("password", urllib.parse.unquote(password))
-        .option("driver", driver)
-        .load()
+            .option("url", url)
+            .option("query", query)
+            .option("user", user)
+            .option("password", urllib.parse.unquote(password))
+            .option("driver", driver)
+            .load()
     )
 
     end = timer()
@@ -252,16 +328,16 @@ def __read_query(
     return ReadResponse(
         df=df,
         execution_statistics=to_execution_statistics(start, end),
-        context={"query": query,},
+        context={"query": query, },
     )
 
 
 def read_all_fn(
-    ctx: typing.Any,
-    r: ReadAllRequest,
-    res: ResolveResourcePathResponse,
-    url: str,
-    driver: str,
+        ctx: typing.Any,
+        r: ReadAllRequest,
+        res: ResolveResourcePathResponse,
+        url: str,
+        driver: str,
 ) -> ReadResponse:
     return __read_query(
         spark=ctx,
@@ -274,13 +350,13 @@ def read_all_fn(
 
 
 def read_fn(
-    ctx: typing.Any,
-    r: ReadRequest,
-    res: ResolveResourceIntentResponse,
-    dialect,
-    url: str,
-    driver: str,
-    store_helper: DataStoreHelper,
+        ctx: typing.Any,
+        r: ReadRequest,
+        res: ResolveResourceIntentResponse,
+        dialect,
+        url: str,
+        driver: str,
+        store_helper: DataStoreHelper,
 ) -> ReadResponse:
     return __read_query(
         spark=ctx,
