@@ -115,9 +115,7 @@ class Asset:
         )
 
         if self.valid_start_dt is None:
-            self.valid_start_dt = (
-                self.effective_end_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
-            )
+            self.valid_start_dt = self.effective_end_dt
 
         if self.valid_end_dt is None:
             self.valid_end_dt = datetime.datetime(year=3000, month=12, day=31)
